# External dependencies
from transformers.trainer_utils import set_seed
set_seed(42)

from pathlib import Path
import numpy as np
import matplotlib.pyplot as plt
import torch
from pyro import distributions as dist
import pyro.distributions as dist
from pykalman import KalmanFilter
from tqdm import tqdm

# Internal dependencies
import sys
sys.path.append('../../')
from src.utils import read_tensor_data, get_array, plot_curves, show
from src.basis_learner import BasisLearner, plot_basis_fns
from src.priors import IGMinnesota, UncorrelatedNormalNNPrior
from src.forecast_utils import get_metrics, export_avg, export_table, save_forecast_results

self_path = Path(__file__).absolute().parent
data_path = self_path / '../../data/processed/curve_data.txt'

if __name__ == '__main__':
    # Device config ("cpu" or "cuda")
    device = 'cpu'

    # Experiment constants
    
    ## B3 Interpoled Yield Curve data
    yield_data = read_tensor_data(str(data_path))
    
    ## Maturities in biz days
    maturities = [21, 42, 63, 126, 252, 504, 756, 1008, 1260, 1512, 1764]

    ## Forecast horizons
    horizons = [5, 20, 60, 120]

    ## Train-test split
    train_start, train_end, test_end = [0, 2833, 3589]

    ## Prior distributions
    minnesota_prior = IGMinnesota(IG_a=1e-1, IG_b=1e-3, minnesota_lambda=0.5, 
                                  minnesota_gamma=0.9, device=device)
    nn_prior = UncorrelatedNormalNNPrior(nn_prior_variance=0.05)

    # Torch tensor conversion.
    tr_yield, te_yield = yield_data[train_start:train_end], yield_data[train_end:test_end]
    tr_y, te_y = [torch.tensor(series[:, maturities], dtype=torch.float64).to(device)
                  for series in [tr_yield, te_yield]]
    
    # Normalization (zero mean) functions
    def normalize(x):
        global tr_y
        
        tr_y_bar = tr_y.mean(axis=0)
        if torch.is_tensor(x):
            return x - tr_y_bar
        return x - get_array(tr_y_bar)

    def denormalize(x):
        global tr_y

        tr_y_bar = tr_y.mean(axis=0)
        if torch.is_tensor(x):
            return x + tr_y_bar
        return x + get_array(tr_y_bar)
    
    # Demeaned tr_y
    tr_y_norm = normalize(tr_y)

    # Maturity torch tensor (in biz years)
    m = torch.tensor(maturities, dtype=torch.float64)[..., None]/252
    m = m.to(device)

    # The Neural network is initialized with a model that reproduces (in the squared error sense)
    # the DSV model Factor Loading Matrix
    basis_learner = BasisLearner(nb_factors=4).double().to(device)
    basis_learner.load_weights(self_path / 'starting_factor_loadings/blank_basis_learner_4.pk')
    
    # Model definition
    init_dist_mu = torch.tensor([0.], requires_grad=True, dtype=torch.float64, device=device)
    init_dist_sigma = torch.tensor([1.], requires_grad=True, dtype=torch.float64, device=device)
    init_dist = dist.Normal(init_dist_mu, init_dist_sigma).expand([4]).to_event()

    # Transition matrix
    trans_matrix = torch.tensor(np.eye(4), requires_grad=True, dtype=torch.float64, device=device)

    # E(w_t)
    mu_trans = torch.tensor(np.random.randn(4)*0, requires_grad=False, dtype=torch.float64, device=device)

    # State covariance matrix
    trans_sigma = torch.tensor([1e-3]*(4), requires_grad=True, dtype=torch.float64, device=device)

    # Noise distribution (w_t)
    trans_dist = dist.Normal(mu_trans, trans_sigma).expand([4]).to_event()

    # Observation matrix (H_\theta)
    observation_matrix = basis_learner(m).T
    observation_sigma = torch.tensor([5e-4]*len(maturities), dtype=torch.float64, device=device)
    
    # Observation noise 
    observation_dist = dist.Normal(0., observation_sigma).expand([len(maturities)]).to_event(1)

    # Linear Gaussian State Space model
    g = dist.GaussianHMM(initial_dist=init_dist, transition_matrix=trans_matrix, transition_dist=trans_dist, observation_matrix=observation_matrix, 
                         observation_dist=observation_dist, duration=train_end-train_start)

    # Optimizer settings
    opt = torch.optim.Adam([
        {'params': list(basis_learner.parameters()), 'lr': 1e-3},
        {'params': [trans_sigma, init_dist_sigma], 'lr': 1e-4},
        {'params': [observation_sigma], 'lr': 5e-5},
        {'params': [trans_matrix, mu_trans, init_dist_mu], 'lr': 1e-3},
    ])

   # Estimation iterative proccess
    for i in range(500):
        # Zera gradiente
        opt.zero_grad()
        
        # Obtém nova base de funções
        observation_matrix = basis_learner(m).T
        
        # Define modelo de espaço de estados
        g = dist.GaussianHMM(initial_dist=init_dist, transition_matrix=trans_matrix, transition_dist=trans_dist,
                            observation_matrix=observation_matrix, observation_dist=observation_dist, duration=train_end-train_start)
        
        # Negative log likelihood: -p(theta|y) \propto - (p(theta) + p(y| theta))
        loss = -minnesota_prior.log_pdf(trans_sigma, trans_matrix) - nn_prior.log_pdf(basis_learner)
        loss = loss - g.log_prob(tr_y_norm)

        # Backpropagation
        loss.backward(retain_graph=True)
        
        # Gradient update step
        opt.step()
        
        # Logging
        if (i+1) % 100 == 0:
            print(f'{i}/{500} : {loss.item():,}')
            print(np.round(get_array(trans_sigma), 4))
            print(np.round(get_array(observation_sigma), 4))
            print('-'*10)
        
        # Clamps negative covariance values before next iteration
        with torch.no_grad():
            observation_sigma += (observation_sigma < 0)*(-1*observation_sigma.min() + 1e-7)
            trans_sigma += (trans_sigma < 0)*(-1*trans_sigma.min() + 1e-7)
    
    # Print estimated parameters
    print(f'Estimated transition_sigma = {np.round(get_array(trans_sigma), 4)}')
    print(f'Estimated observation_sigma = {np.round(get_array(observation_sigma), 4)}')
    print(f'Estimated transition_matrix = {np.round(get_array(trans_matrix), 3)}')

    # Factor loading function plot
    plot_basis_fns(basis_learner, device=device, title='Estimated Factor Loadings')

    # Kalman Filter
    kf = KalmanFilter(transition_matrices=trans_matrix.detach().numpy(),
                      observation_matrices=observation_matrix.detach().numpy().T,
                      transition_covariance=np.diag(trans_sigma.detach().numpy()),
                      observation_covariance=np.diag(observation_sigma.detach().numpy()),
                      initial_state_mean=get_array(g.filter(tr_y_norm[0:1]).mean),
                      initial_state_covariance=np.diag(get_array(g.filter(tr_y_norm[0:1]).variance)))

    filter_states, filter_covs = kf.filter(get_array(tr_y_norm))

    # Forecast
    def get_curve_from_factors(basis_learner, factors):
        assert len(factors.shape) > 1
        
        with torch.no_grad():
            observation_matrix = basis_learner(m).T
        
        return denormalize(factors@get_array(observation_matrix))

    def forecast(factors, transition_matrix, h=1):
        trans_matrix_np = get_array(trans_matrix)
        last_pred = factors[-1:]
        predictions = [None]*h
        
        for i in range(h):
            last_pred = last_pred@trans_matrix_np
            predictions[i] = last_pred
        
        return np.concatenate(predictions)
    
    def recursive_forecast(train_tensor, test_tensor, horizons, window_step=1):
        global kf
        global trans_matrix
        global basis_learner
        
        forecast_results = {h: {} for h in horizons}
        print('Forecasting...')

        for h in tqdm(horizons):
            current_state, current_cov = filter_states[-1], filter_covs[-1]
            
            for i in np.arange(0, test_tensor.shape[0]-h, window_step):
                current_index = i
                forecast_results[h][i] = {}
                
                current_dataset = torch.cat([train_tensor, normalize(test_tensor)[0:current_index]])
                
                # Saves the Current Index and the real t+h yield curve
                forecast_results[h][i]['current_index'] = current_index
                forecast_results[h][i]['real'] = get_array(test_tensor[current_index+h-1])
                
                # Random walk benchmark
                forecast_results[h][i]['rw'] = denormalize(get_array(current_dataset))[-1]
                
                with torch.no_grad():
                    # Updates the Kalman Filter using y_t
                    current_state, current_cov = kf.filter_update(current_state, current_cov,
                                                                  get_array(current_dataset[-1]))
                    current_state, current_cov = np.array(current_state), np.array(current_cov)
                    
                    # Forecasts the factor vector using the estimated transition matrix
                    new_state = forecast(current_state.reshape(1, -1), trans_matrix, h)
                    
                    # Reconstructs the t+h yield curve from the forecasted factors
                    forecasted = get_curve_from_factors(basis_learner, new_state[-1:]).squeeze()
                
                forecast_results[h][i]['factor_forecast'] = new_state
                forecast_results[h][i]['forecast'] = forecasted
        
        return forecast_results
    
    def get_metrics(forecast_results, horizons):
        forecast_metrics = {h: {} for h in horizons}
        
        for horizon in horizons:
            n = len(forecast_results[horizon])
            rw_mse, mse, mae = 0., 0., 0.
            
            for dict_step in forecast_results[horizon].values():
                mse += (dict_step['real'] - dict_step['forecast'])**2
                mae += np.abs(dict_step['real'] - dict_step['forecast'])
                rw_mse += (dict_step['real'] - dict_step['rw'])**2

            forecast_metrics[horizon]['mae'] = (mae/n)
            forecast_metrics[horizon]['mse'] = (mse/n)
            forecast_metrics[horizon]['rw_mse'] = (rw_mse/n)
            forecast_metrics[horizon]['rw_rmse'] = np.sqrt(rw_mse/n)
            forecast_metrics[horizon]['rmse'] = (np.sqrt(mse/n))
        return forecast_metrics

    def export_table(mse_list):
        lista = np.round(np.sqrt(mse_list)*1e3, 2)
        print(' & '.join([str(x) for x in lista]))

    def export_avg(mse_list):
        lista = np.round(np.sqrt(mse_list)*1e3, 2)
        return np.mean(lista)

    forecast_results = recursive_forecast(tr_y_norm, te_y, [5, 20, 60, 120])
    forecast_metrics = get_metrics(forecast_results, [5, 20, 60, 120])

    print('Average out-of-sample RMSE (times 10^3)')
    
    for h in horizons:
        print(f'(h = {h}) {export_avg(forecast_metrics[h]["mse"])}')

    save_forecast_results(forecast_metrics, self_path / '../../reports/forecasting_benchmark/4nnss.json')
    print(f'Full forecast report saved on reports/forecasting_benchmark/4nnss.json')
    